#include <stdio.h>
#include <string.h>
#include "pgp.h"
#include "language.h"
#include "usuals.h"  /* typedefs for byte, word16, boolean, etc. */
#include "hash.h"
#include "fileio.h"
#include "xnew.h"


static int _Hashfile0_len(struct HashContext *ctx, FILE * f, word32 longcount);
static int _Hashfile0(struct HashContext *ctx, FILE * inFile);
static int _Hashfile(struct HashContext *ctx, char *filename);
static void _Hash_addbuffer(struct HashContext *ctx, byte * buf, int buflen, byte digest[]);


int HashSize(int Hash)
{
    if (Hash == MD5_ALGORITHM_BYTE)
        return 16;
    else if (Hash == SHA1_ALGORITHM_BYTE)
        return 20;
    else if (Hash == RIPEMD160_ALGORITHM_BYTE)
        return 20;
#ifdef ENABLE_SHA2
    else if (Hash == SHA256_ALGORITHM_BYTE)
        return 32;
    else if (Hash == SHA384_ALGORITHM_BYTE)
        return 48;
    else if (Hash == SHA512_ALGORITHM_BYTE)
        return 64;
#endif
#ifdef ENABLE_TIGER
    else if (Hash == TIGER192_ALGORITHM_BYTE)
        return 24;
#endif
#ifdef ENABLE_HAVAL
    else if (Hash == HAVAL_5_160_ALGORITHM_BYTE)
        return 20;
    else if (Hash == HAVAL_5_256_ALGORITHM_BYTE)
        return 32;
#endif
#ifdef ENABLE_DOUBLESHA
    else if (Hash == SHA1X_ALGORITHM_BYTE)
        return 40;
#endif
    return 0;
    /* MUST return 0 for unsupported hashes !
       because HashSize() is also used to verify if hash algo is supported */
}

static char szMD5[]="MD5";
static char szSHA1[]="SHA1";
static char szRIPEMD160[]="RIPEMD160";
static char szSHA256[]="SHA256";
static char szSHA384[]="SHA384";
static char szSHA512[]="SHA512";
static char szTIGER192[]="TIGER192";
static char szHAVAL_5_160[]="HAVAL-5-160";
static char szHAVAL_5_256[]="HAVAL-5-256";
static char szSHA1X[]="SHA1x";
static char szOther[12]="";

char * HashName(int Hash)
{
    if (Hash == MD5_ALGORITHM_BYTE)
        return szMD5;
    else if (Hash == SHA1_ALGORITHM_BYTE)
        return szSHA1;
    else if (Hash == RIPEMD160_ALGORITHM_BYTE)
        return szRIPEMD160;
    else if (Hash == SHA256_ALGORITHM_BYTE)
        return szSHA256;
    else if (Hash == SHA384_ALGORITHM_BYTE)
        return szSHA384;
    else if (Hash == SHA512_ALGORITHM_BYTE)
        return szSHA512;
    else if (Hash == TIGER192_ALGORITHM_BYTE)
        return szTIGER192;
    else if (Hash == HAVAL_5_160_ALGORITHM_BYTE)
        return szHAVAL_5_160;
    else if (Hash == HAVAL_5_256_ALGORITHM_BYTE)
        return szHAVAL_5_256;
    else if (Hash == SHA1X_ALGORITHM_BYTE)
        return szSHA1X;
    sprintf(szOther, "%d", Hash);
    return szOther;
}


void HashInit(struct HashContext *context, int Hash)
{
    context->Hash = Hash;
    context->hashsize = HashSize(context->Hash);
    if (context->Hash == MD5_ALGORITHM_BYTE)
        MD5Init(&context->ctx.md5ctx);
    else if (context->Hash == SHA1_ALGORITHM_BYTE)
        SHA1Init(&context->ctx.sha1ctx);
    else if (context->Hash == RIPEMD160_ALGORITHM_BYTE)
        RMDinit(&context->ctx.rmdctx);
#ifdef ENABLE_SHA2
    else if (context->Hash == SHA256_ALGORITHM_BYTE)
        sha_init(&context->ctx.sha256ctx);
    else if (context->Hash == SHA384_ALGORITHM_BYTE)
        sha_init384(&context->ctx.sha384ctx);
    else if (context->Hash == SHA512_ALGORITHM_BYTE)
        sha_init512(&context->ctx.sha512ctx);
#endif
#ifdef ENABLE_TIGER
    else if (context->Hash == TIGER192_ALGORITHM_BYTE)
        tigerinit(&context->ctx.tctx);
#endif
#ifdef ENABLE_HAVAL
    else if (context->Hash == HAVAL_5_160_ALGORITHM_BYTE) {
        context->ctx.havalst.pass = 5;
        context->ctx.havalst.fptlen = 160;
        haval_start(&context->ctx.havalst);
    }
    else if (context->Hash == HAVAL_5_256_ALGORITHM_BYTE) {
        context->ctx.havalst.pass = 5;
        context->ctx.havalst.fptlen = 256;
        haval_start(&context->ctx.havalst);
    }
#endif
#ifdef ENABLE_DOUBLESHA
    else if (context->Hash == SHA1X_ALGORITHM_BYTE)
        shaDoubleInit(&context->ctx.sha1x);
#endif
}

void HashUpdate(struct HashContext *context, unsigned char const *buf, unsigned len)
{
    if (context->Hash == MD5_ALGORITHM_BYTE)
        MD5Update(&context->ctx.md5ctx, buf, len);
    else if (context->Hash == SHA1_ALGORITHM_BYTE)
        SHA1Update(&context->ctx.sha1ctx, buf, len);
    else if (context->Hash == RIPEMD160_ALGORITHM_BYTE)
        RMDUpdate(&context->ctx.rmdctx, buf, len);
#ifdef ENABLE_SHA2
    else if (context->Hash == SHA256_ALGORITHM_BYTE)
        sha_process(&context->ctx.sha256ctx, (char *)buf, len);
    else if (context->Hash == SHA384_ALGORITHM_BYTE)
        sha_process384(&context->ctx.sha384ctx, (char *)buf, len);
    else if (context->Hash == SHA512_ALGORITHM_BYTE)
        sha_process512(&context->ctx.sha512ctx, (char *)buf, len);
#endif
#ifdef ENABLE_TIGER
    else if (context->Hash == TIGER192_ALGORITHM_BYTE)
        tigerUpdate(&context->ctx.tctx, buf, len);
#endif
#ifdef ENABLE_HAVAL
    else if (context->Hash == HAVAL_5_160_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_256_ALGORITHM_BYTE)
        haval_hash(&context->ctx.havalst, (char *)buf, len);
#endif
#ifdef ENABLE_DOUBLESHA
    else if (context->Hash == SHA1X_ALGORITHM_BYTE)
        shaDoubleUpdate(&context->ctx.sha1x, buf, len);
#endif
}

void HashFinal(byte *buf, struct HashContext *context)
{
    if (context->Hash == MD5_ALGORITHM_BYTE)
        MD5Final(buf, &context->ctx.md5ctx);
    else if (context->Hash == SHA1_ALGORITHM_BYTE)
        SHA1Final(buf, &context->ctx.sha1ctx);
    else if (context->Hash == RIPEMD160_ALGORITHM_BYTE)
        RMDFinal(buf, &context->ctx.rmdctx);
#ifdef ENABLE_SHA2
    else if (context->Hash == SHA256_ALGORITHM_BYTE)
        sha_done(&context->ctx.sha256ctx, buf);
    else if (context->Hash == SHA384_ALGORITHM_BYTE)
        sha_done384(&context->ctx.sha384ctx, buf);
    else if (context->Hash == SHA512_ALGORITHM_BYTE)
        sha_done512(&context->ctx.sha512ctx, buf);
#endif
#ifdef ENABLE_TIGER
    else if (context->Hash == TIGER192_ALGORITHM_BYTE) {
        int j; byte b[24];
        tigerFinal(b, &context->ctx.tctx);
        /* damn... GnuPG have different byte order */
        for(j=0; j<24; j++) buf[j^7] = b[j];
    }
#endif
#ifdef ENABLE_HAVAL
    else if (context->Hash == HAVAL_5_160_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_256_ALGORITHM_BYTE)
        haval_end(&context->ctx.havalst, buf);
#endif
#ifdef ENABLE_DOUBLESHA
    else if (context->Hash == SHA1X_ALGORITHM_BYTE)
        shaDoubleFinal(&context->ctx.sha1x, buf);
#endif
}



int Hashfile0_len(struct HashContext *context, FILE *f, word32 longcount, int Hash)
{
    context->Hash = Hash;
    context->hashsize = HashSize(context->Hash);
    if (context->hashsize == 0)
        return -1;
    /*
    if (context->Hash == MD5_ALGORITHM_BYTE)
        return MDfile0_len(&context->ctx.md5ctx, f, longcount);
    else if (context->Hash == SHA1_ALGORITHM_BYTE)
        return SHA1file0_len(&context->ctx.sha1ctx, f, longcount);
    else if (context->Hash == RIPEMD160_ALGORITHM_BYTE)
        return RMDfile0_len(&context->ctx.rmdctx, f, longcount);
    else if (context->Hash == SHA256_ALGORITHM_BYTE ||
             context->Hash == SHA384_ALGORITHM_BYTE ||
             context->Hash == SHA512_ALGORITHM_BYTE ||
             context->Hash == TIGER192_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_160_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_256_ALGORITHM_BYTE ||
             context->Hash == SHA1X_ALGORITHM_BYTE)
    */
        return _Hashfile0_len(context, f, longcount);
    return -1;
}

int Hashfile0(struct HashContext *context, FILE * inFile, int Hash)
{
    context->Hash = Hash;
    context->hashsize = HashSize(context->Hash);
    return _Hashfile0(context, inFile);
}

// Computes the message digest for a specified file
int Hashfile(struct HashContext *context, char *filename, int Hash)
{
    context->Hash = Hash;
    context->hashsize = HashSize(context->Hash);
    if (context->hashsize == 0)
        return -1;
    /*
    if (context->Hash == MD5_ALGORITHM_BYTE)
        return MDfile(&context->ctx.md5ctx, filename);
    else if (context->Hash == SHA1_ALGORITHM_BYTE)
        return SHA1file(&context->ctx.sha1ctx, filename);
    else if (context->Hash == RIPEMD160_ALGORITHM_BYTE)
        return RMDfile(&context->ctx.rmdctx, filename);
    else if (context->Hash == SHA256_ALGORITHM_BYTE ||
             context->Hash == SHA384_ALGORITHM_BYTE ||
             context->Hash == SHA512_ALGORITHM_BYTE ||
             context->Hash == TIGER192_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_160_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_256_ALGORITHM_BYTE ||
             context->Hash == SHA1X_ALGORITHM_BYTE)
    */
        return _Hashfile(context, filename);
    return -1;
}

// Used to finish the message digest, including extra material.
void Hash_addbuffer (struct HashContext *context, byte *buf, int buflen, byte *digest)
{
    /*
    if (context->Hash == MD5_ALGORITHM_BYTE)
        MD_addbuffer(&context->ctx.md5ctx, buf, buflen, digest);
    else if (context->Hash == SHA1_ALGORITHM_BYTE)
        SHA1_addbuffer(&context->ctx.sha1ctx, buf, buflen, digest);
    else if (context->Hash == RIPEMD160_ALGORITHM_BYTE)
        RMD_addbuffer(&context->ctx.rmdctx, buf, buflen, digest);
    else if (context->Hash == SHA256_ALGORITHM_BYTE ||
             context->Hash == SHA384_ALGORITHM_BYTE ||
             context->Hash == SHA512_ALGORITHM_BYTE ||
             context->Hash == TIGER192_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_160_ALGORITHM_BYTE ||
             context->Hash == HAVAL_5_256_ALGORITHM_BYTE ||
             context->Hash == SHA1X_ALGORITHM_BYTE)
    */
        _Hash_addbuffer(context, buf, buflen, digest);
}


//----------------------------
/* Note - the routines in this module, except for Hash_addbuffer,
 * do not "finish" the Hash calculation.  Hash_addbuffer finishes the
 * calculation in each case, usually to append the timestamp and class info.
 */

/* Computes the message digest for a file from current position for
   longcount bytes. */


static int _Hashfile0_len(struct HashContext *ctx, FILE * f, word32 longcount)
{
    int bytecount;
    XNEW(buffer,unsigned char,1024);

    HashInit(ctx, ctx->Hash);
    /* Process 1024 bytes at a time... */
    do {
        if (longcount < (word32) 1024)
            bytecount = (int) longcount;
        else
            bytecount = 1024;
        bytecount = fread(buffer, 1, bytecount, f);
        if (bytecount > 0) {
            HashUpdate(ctx, buffer, bytecount);
            longcount -= bytecount;
#ifdef MACPGP
                mac_poll_for_break();
#endif
        }
        /* if text block was short, exit loop */
    } while (bytecount == 1024);
    XDEL(buffer);
    return 0;
}                               /* Hashfile0_len */


/* Computes the message digest for a file from current position to EOF. */

static int _Hashfile0(struct HashContext *ctx, FILE * inFile)
{
    int bytes;
    XNEW(buffer,unsigned char,1024);

    HashInit(ctx, ctx->Hash);
    while ((bytes = fread(buffer, 1, 1024, inFile)) != 0)
#ifdef MACPGP
                {
                mac_poll_for_break();
                HashUpdate(MDbuf,buffer,bytes);
                }
#else
        HashUpdate(ctx, buffer, bytes);
#endif
    XDEL(buffer);
    return 0;
}

/* Computes the message digest for a specified file */

static int _Hashfile(struct HashContext *ctx, char *filename)
{
    FILE *inFile;
    inFile = fopen(filename, FOPRBIN);

    if (inFile == NULL) {
        fprintf(pgpout, LANG("\n\007Can't open file '%s'\n"), filename);
        return -1;
    }
    _Hashfile0(ctx, inFile);
    fclose(inFile);
    return 0;
}

/* Add a buffer's worth of data to the Hash computation.  If a digest
 * pointer is supplied, complete the computation and write the digest.
 */
static void _Hash_addbuffer(struct HashContext *ctx, byte * buf, int buflen, byte digest[])
{
    HashUpdate(ctx, buf, buflen);
    if (digest) {
        HashFinal(digest, ctx);
        burn(*ctx);       /* Paranoia */
    }
}
